import os
import numpy as np

from phi.physics.field import CenteredGrid, StaggeredGrid
from phi.physics.field.staggered_grid import stack_staggered_components


# Views
FRONT = 'front'
RIGHT = 'right'
TOP = 'top'

# Vector display
LENGTH = 'length'
VECTOR2 = 'vec2'


class PlotlyFigureBuilder(object):

    def __init__(self,
                 batches=slice(None),
                 depths=slice(None),
                 antisymmetry=False,
                 view=FRONT,
                 component=LENGTH,
                 draw_arrows_backward=True,
                 max_vector_resolution=18,
                 max_resolution=128):
        self.batches = batches
        self.depths = depths
        self.antisymmetry = antisymmetry
        self.view = view
        self.component = component
        self.draw_arrows_backward = draw_arrows_backward
        self.max_vector_resolution = max_vector_resolution
        self.max_resolution = max_resolution

    def select_batch(self, batch):
        if batch is None:
            self.batches = slice(None)
        elif isinstance(batch, int):
            self.batches = [batch]
        else:
            self.batches = batch

    def select_depth(self, depth):
        if depth is None:
            self.depths = slice(None)
        elif isinstance(depth, int):
            self.depths = [depth]
        else:
            self.depths = depth

    def save_figures(self, directory, fieldname, time, data, same_scale_data=None):
        import matplotlib.pyplot as plt
        batches = self.batches if self.batches is not None else range(data.shape[0])
        for batch in batches:
            if len(data.shape) == 5:
                for depth in self.get_selected_slices(data.shape):
                    path = os.path.join(directory, '%s_batch%04d_depth%04d_%04d.png' % (fieldname, batch, depth, time))
                    fig = self.create_figure(data, batch=batch, depth=depth, same_scale_data=same_scale_data, library='matplotlib')
                    plt.savefig(path)
                    plt.close()
                    yield path
            else:
                path = os.path.join(directory, '%s_batch%04d_%04d.png' % (fieldname, batch, time))
                fig = self.create_figure(data, batch=batch, same_scale_data=same_scale_data, library='matplotlib')
                plt.savefig(path)
                plt.close()
                yield path

    def get_selected_slices(self, shape):
        try:
            selected_depths = np.arange(self.slice_count(shape))[self.depths]
        except:
            selected_depths = [self.slice_count(shape) - 1]
        return selected_depths

    def slice_dims(self, data):
        if isinstance(data, CenteredGrid):
            return data.data.shape
        if isinstance(data, StaggeredGrid):
            return data.data[0].data.shape
        else:
            return data.shape

    def create_figure(self, data, same_scale_data=None, batch=None, depth=None, library='matplotlib'):
        shape = self.slice_dims(data)
        # Determine batch
        if shape[0] == 1:
            batch = 0
        if batch is None:
            try:
                selected_batches = np.arange(shape[0])[self.batches]
                if len(selected_batches) != 1:
                    raise ValueError('no batch specified and default batches contains more than one element')
            except:
                return None
            batch = selected_batches[0]
        # Determine slice
        if depth is None and len(shape) == 5:
            selected_depths = self.get_selected_slices(shape)
            if len(selected_depths) != 1:
                raise ValueError('no depth specified and default depths contains more than one element')
            depth = selected_depths[0]

        # special handling for staggered grids and anti symmetry
        if isinstance(data, StaggeredGrid):
            if not self.antisymmetry:
                data = data.staggered_tensor()
            else:
                dims = len(data.data[0].data.shape) - 2  # any better way to get this?
                dataxyz = []
                for i in range(dims):
                    c = data.data[i].data
                    factor = -1. if i == (dims - 1) else 1.  # add (instead of subtract) for X dim
                    cdiff = c[..., ::-1,0:1] - (c[...,0:1] * factor)
                    dataxyz.append(cdiff)
                data = stack_staggered_components(dataxyz)
            shape = data.shape
        else:
            if isinstance(data, CenteredGrid):
                data = data.data
            if self.antisymmetry:
                data = data - data[..., ::-1, :]

        # Select batch
        if batch < shape[0]:
            data = data[batch, ...]
        else:
            return {'data': [{'type': 'heatmap', 'z': [[0]]}]}

        if np.issubdtype(data.dtype, np.complex):
            data = np.real(data)

        # 1D graph
        if len(shape) == 3:
            return self.graphs(data, library)

        # 3D projection / Select depth
        if len(shape) == 5:
            if self.view == FRONT:
                data = data[:, min(depth, shape[2]), :, :]
            elif self.view == RIGHT:
                data = data[:, :, min(depth, shape[3]), :]
                data = np.transpose(data, axes=(1, 0, 2))
            elif self.view == TOP:
                data = data[min(depth, shape[1]), :, :, :]
            else:
                data = data[0, ...]

        # Create figure
        component = 0 if shape[-1] == 1 else self.component

        if component == VECTOR2:
            # Downsample
            while np.prod(data.shape[:-1]) > self.max_vector_resolution ** 2:
                data = data[::2, ::2, :] * 0.5
            data = data[..., ::-1]
            return self.draw_vector_field(data, library)

        elif component == LENGTH:
            if shape[-1] == 3:
                data = np.sqrt(data[..., 0:1]**2 + data[..., 1:2]**2 + data[..., 2:3]**2)
            else:
                data = np.sqrt(data[..., 0:1]**2 + data[..., 1:2]**2)
        else:
            # Single vector component
            if component >= shape[-1]:
                data = np.zeros_like(data[..., 0:1])
            else:
                data = data[..., shape[-1] - 1 - component:shape[-1] - component]

        # Downsample
        while np.prod(data.shape[:-1]) > self.max_resolution ** 2:
            data = data[::2, ::2, :]
        if same_scale_data is not None:
            return self.heatmap(data[..., 0], library, minmax=global_minmax(same_scale_data))
        else:
            return self.heatmap(data[..., 0], library)

    def slice_count(self, shape):
        if len(shape) <= 4:
            return 1
        if self.view == FRONT:
            return shape[1]
        elif self.view == TOP:
            return shape[2]
        elif self.view == RIGHT:
            return shape[3]
        else:
            raise ValueError('Illegal view: %s' % self.view)

    def heatmap(self, z, library, minmax=None):
        if library == 'dash':
            args = {'z': z, 'type': 'heatmap'}
            if minmax is not None:
                args['zmin'] = minmax[0]
                args['zmax'] = minmax[1]
            return {'data': [args]}
        elif library == 'matplotlib':
            import matplotlib.pyplot as plt
            fig = plt.figure()
            plt.imshow(z, cmap='bwr', origin='lower')
            return fig
        else:
            raise NotImplementedError()

    def graphs(self, data, library):
        x = np.arange(data.shape[0])
        if library == 'dash':
            graphs = [{'mode': 'markers+lines', 'type': 'scatter', 'x': x, 'y': data[:, i]} for i in range(data.shape[-1])]
            return {'data': graphs}
        else:
            import matplotlib.pyplot as plt
            fig = plt.figure()
            for i in range(data.shape[-1]):
                plt.plot(x, data[:,i])
            return fig

    def draw_vector_field(self, vector_field, library):
        import plotly.figure_factory as ff
        x, y = np.meshgrid(np.arange(0, vector_field.shape[1], 1), np.arange(0, vector_field.shape[0], 1))
        if library == 'dash':
            if self.draw_arrows_backward:
                return ff.create_quiver(x - vector_field[..., 0], y - vector_field[..., 1], vector_field[..., 0], vector_field[..., 1], scale=1.0)
            else:
                return ff.create_quiver(x, y, vector_field[..., 0], vector_field[..., 1], scale=1.0)
        else:
            raise NotImplementedError()

    def empty_figure(self, library):
        if library == 'dash':
            return {
                'data': [{'z': None, 'type': 'heatmap'}]
            }
        elif library == 'matplotlib':
            import matplotlib.pyplot as plt
            fig = plt.figure()
            return fig


def global_minmax(arrays):
    global_min = np.minimum(*[np.min(data) for data in arrays])
    global_max = np.maximum(*[np.max(data) for data in arrays])
    return global_min, global_max



import numpy as np

from phi import struct
from phi.physics.world import World
from phi.physics import Physics
from phi.physics.collective import CollectivePhysics
from phi.struct import VARIABLES
from phi.struct.functions import mappable
from .util import placeholder


def tf_bake_graph(world, session):
    # --- Build placeholder state ---
    with VARIABLES:
        state_in = placeholder(world.state.staticshape, dtype=world.state.dtype)
    dt = placeholder(())
    # --- Build graph ---
    state_out = world.physics.step(state_in, dt=dt)
    world.physics = BakedWorldPhysics(world.physics, session, state_in, state_out, dt, world=world)
    for name, sysstate in world.state.states.items():
        sysstate_in = state_in[name]
        sysstate_out = state_out[name]
        baked_physics = BakedPhysics(session, sysstate_in, sysstate_out, dt)
        world.physics.add(name, baked_physics)


def tf_bake_subgraph(tracker, session):
    tfworld = World()
    tfworld.add(tracker.state)
    # --- Build placeholder state ---
    with VARIABLES:
        state_in = placeholder(tracker.state.staticshape, dtype=tracker.state.dtype)
    dt = placeholder(())
    # --- Build graph ---
    state_out = tracker.world.physics.substep(state_in, tracker.world.state, dt)
    tracker.physics = BakedPhysics(session, state_in, state_out, dt)
    return tfworld


class BakedPhysics(Physics):

    def __init__(self, session, state_in, state_out, dt):
        Physics.__init__(self, {})
        self.state_in = state_in
        self.state_out = state_out
        self.session = session
        self.dt = dt

    def step(self, state, dt=1.0, **dependent_states):
        for key, value in dependent_states:
            assert not value, 'Baked subgraph can only be executed without dependencies'
        return self.session.run(self.state_out, {self.state_in: state, self.dt: dt})


class BakedWorldPhysics(CollectivePhysics):

    def __init__(self, physics, session, state_in, state_out, dt, world):
        CollectivePhysics.__init__(self)
        self._physics = physics.physics
        self.state_in = state_in
        self.state_out = state_out
        self.session = session
        self.dt = dt
        self.world = world

    def step(self, world_state, dt=1.0, **dependent_states):
        return self.run(self.state_out, dt, world_state)

    def run(self, fetches, dt=1.0, world_state=None):
        if world_state is None:
            world_state = self.world.state
        feed_dict = {self.state_in: world_state, self.dt: dt}
        return self.session.run(fetches, feed_dict)



import warnings

from phi import struct, math
from phi.geom import GLOBAL_AXIS_ORDER
from ._geom import Geometry
from ._sphere import Sphere


@struct.definition()
class RotatedGeometry(Geometry):

    def __init__(self, geometry, angle, **kwargs):
        Geometry.__init__(self, **struct.kwargs(locals()))

    @struct.constant()
    def geometry(self, geometry):
        assert isinstance(geometry, Geometry)
        if isinstance(geometry, RotatedGeometry):
            warnings.warn('Using RotatedGeometry of RotatedGeometry. Consider simplifying your setup.')
        return geometry

    @struct.constant()
    def angle(self, rotation):
        return rotation

    @property
    def center(self):
        return self.geometry.center

    def global_to_child(self, location):
        """ Inverse transform """
        delta = location - self.center
        if math.staticshape(location)[-1] == 2:
            angle = -math.batch_align(self.angle, 0, location)
            sin = math.sin(angle)
            cos = math.cos(angle)
            y, x = math.unstack(delta, axis=-1)
            if GLOBAL_AXIS_ORDER.is_x_first:
                x, y = y, x
            rot_x = cos * x - sin * y
            rot_y = sin * x + cos * y
            rotated = math.stack([rot_y, rot_x], axis=-1)
        elif math.staticshape(location)[-1] == 3:
            angle = math.batch_align(self.angle, 1, location)
            raise NotImplementedError('not yet implemented')  # ToDo apply angle
        else:
            raise NotImplementedError('Rotation only supported in 2D and 3D')
        final = rotated + self.center
        return final

    def lies_inside(self, location):
        return self.geometry.lies_inside(self.global_to_child(location))

    def approximate_signed_distance(self, location):
        return self.geometry.approximate_signed_distance(self.global_to_child(location))

    def bounding_radius(self):
        return self.geometry.bounding_radius()

    def bounding_half_extent(self):
        bounding_sphere = Sphere(self.center, self.bounding_radius())
        return bounding_sphere.bounding_half_extent()

    @property
    def rank(self):
        return self.geometry.rank

    def shifted(self, delta):
        return self.copied_with(geometry=self.geometry.shifted(delta))

    def rotated(self, angle):
        return self.copied_with(angle=self.angle + angle)


def rotate(geometry, angle):
    """ package-internal rotation function. Users should use Geometry.rotated() instead. """
    assert isinstance(geometry, Geometry)
    if isinstance(geometry, RotatedGeometry):
        total_rotation = geometry.angle + angle  # ToDo concatenate rotations
        return RotatedGeometry(geometry.geometry, total_rotation)
    else:
        return RotatedGeometry(geometry, angle)



from phi import struct
from phi.geom.geometry import Geometry

from .field import GeometryMask
from .field.effect import FieldEffect
from .material import CLOSED, Material
from .physics import Physics, State


@struct.definition()
class Obstacle(State):

    def __init__(self, geometry, material=CLOSED, velocity=0, tags=('obstacle',), **kwargs):
        State.__init__(self, **struct.kwargs(locals()))

    @struct.constant()
    def geometry(self, geometry):
        assert isinstance(geometry, Geometry)
        return geometry

    @struct.constant(default=CLOSED)
    def material(self, material):
        assert isinstance(material, Material)
        return material

    @struct.constant(default=0)
    def velocity(self, velocity):
        return velocity

    @struct.constant(default=0)
    def angular_velocity(self, av):
        return av

    @struct.derived()
    def is_stationary(self):
        return self.velocity is 0 and self.angular_velocity is 0


class GeometryMovement(Physics):

    def __init__(self, geometry_function):
        Physics.__init__(self)
        self.geometry_at = geometry_function

    def step(self, obj, dt=1.0, **dependent_states):
        next_geometry = self.geometry_at(obj.age + dt)
        h = 1e-2 * dt if dt > 0 else 1e-2
        perturbed_geometry = self.geometry_at(obj.age + dt + h)
        velocity = (perturbed_geometry.center - next_geometry.center) / h
        if isinstance(obj, Obstacle):
            return obj.copied_with(geometry=next_geometry, velocity=velocity, age=obj.age + dt)
        if isinstance(obj, FieldEffect):
            with struct.ALL_ITEMS:
                next_field = struct.map(lambda x: x.copied_with(geometries=next_geometry) if isinstance(x, GeometryMask) else x, obj.field, leaf_condition=lambda x: isinstance(x, GeometryMask))
            return obj.copied_with(field=next_field, age=obj.age + dt)





def generate(struct_name, variables=(), constants=(), others=()):
    items = ''
    parameters = ''
    ignore = []
    other_init = ''
    for attr_name in variables:
        items += """
    @struct.variable()
    def {name}(self, {name}):
        return {name}
""".replace('{name}', attr_name)
        parameters += '%s, ' % (attr_name, )
    for prop_name in constants:
        items += """
    @struct.constant()
    def {name}(self, {name}):
        return {name}
        """.replace('{name}', prop_name)
        parameters += '%s, ' % (prop_name,)
    for other_name in others:
        items += """
    @property
    def {name}(self):
        return self._{name}
        """.replace('{name}', other_name)
        parameters += '%s, ' % (other_name,)
        ignore.append("'%s'" % other_name)
        other_init += '\n        self._{name} = {name}'.replace('{name}', other_name)
    if len(ignore) > 0:
        ignore = ', ignore=[%s]' % ', '.join(ignore)
    else:
        ignore = ''

    return """
from phi import struct


@struct.definition()
class %s(struct.Struct):
    
    def __init__(self, %s**kwargs):
        struct.Struct.__init__(self, **struct.kwargs(locals()%s))%s
    %s
""" % (struct_name, parameters, ignore, other_init, items)



from phi.struct import Trait
from phi.backend.dynamic_backend import DYNAMIC_BACKEND as math


class Batched(Trait):
    """
Structs with this trait can tag items with the keyword 'min_rank', representing the number of innate inner dimensions of tensor values of that item.
All further, outer dimensions are assumed to be batch dimensions.

Example:
    @struct.definition(traits=[math.BATCHED])\n
    class MyStruct(struct.Struct):
        @struct.constant(min_rank=1)\n
        def batched_constant(self, c):
            assert math.ndims(c) >= 1  # Will always be fulfilled\n
            return c

For each item with the 'min_rank' keyword, (1) the specified minimum tensor rank is ensured by expanding the dimensions if necessary and (2) batch shape checks are performed during validation.
Also, all additional dimensions, called `batch dimensions` are collected and cross-checked between all items.
The batch dimensions of a valid batched struct can be accessed as `struct.batch_shape`, and `struct.batch_rank` holds the corresponding length.

When a struct with inconsistent batch dimensions is validated, a `ShapeMismatch` error is raised, typically upon struct creation.

The Batched trait also ensures that all values are converted to tensors before the validation function is called.
    """

    def check_argument(self, struct_class, item, keyword, value):
        assert keyword == 'min_rank'
        assert isinstance(value, int) or callable(value), value

    def endow(self, struct):
        struct.batch_shape = None
        struct.batch_rank = None

    def pre_validate_struct(self, struct):
        struct.batch_shape = None
        struct.batch_rank = None

    def pre_validated(self, struct, item, value):
        tensor = math.as_tensor(value)
        min_rank = item.trait_kwargs['min_rank']
        if callable(min_rank):
            min_rank = min_rank(struct)
        shape = math.staticshape(value)
        if len(shape) < min_rank:
            tensor = math.expand_dims(tensor, axis=0, number=min_rank - len(shape))
            shape = math.staticshape(value)
        batch_shape = shape[:-min_rank if min_rank != 0 else None]
        if struct.batch_shape is None:
            struct.batch_shape = batch_shape
        else:
            struct.batch_shape = _combined_shape(batch_shape, struct.batch_shape, item, struct)
        struct.batch_rank = len(struct.batch_shape)
        return tensor


BATCHED = Batched(keywords=['min_rank'])


def _combined_shape(shape1, shape2, prop, obj):
    rank = max(len(shape1), len(shape2))
    resulting_shape = []
    for i in range(1, rank + 1):
        dim1 = shape1[-i] if len(shape1) >= i else 1
        dim2 = shape2[-i] if len(shape2) >= i else 1
        try:
            resulting_shape.append(_combined_dim(dim1, dim2))
        except AssertionError:
            raise ShapeMismatch("Batch dimension %d with value %d of '%s' of %s does not match other properties with value %d. Occured during comparison of batch shapes %s and %s" % (-i, dim1, prop, obj, dim2, shape1, shape2))
    return tuple(resulting_shape[::-1])


def _combined_dim(dim1, dim2):
    if dim1 is None or dim2 is None:
        return None
    if dim1 == 1:
        return dim2
    if dim2 == 1:
        return dim1
    assert dim1 == dim2
    return dim1


class ShapeMismatch(ValueError):
    """
Raised when a shape check fails, i.e. when tensors that require compatible shapes do not match.
It is a subclass of `ValueError` because ValueErrors are often raised in this case.
    """

    def __init__(self, *args):
        ValueError.__init__(self, *args)
